/*
** http.c - part of GrabURL 2.0
** Copyright (C) 1996-97 Serge Emond
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <sys/time.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include "graburl.h"
#include "config.h"
#include "macros.h"
#include "libgurl.h"
#include "http.h"
#include "log.h"

#define magic_makedirs(a, b) \
	(grab->uri->flags & URI_NODIRS ? 0 : makedirs(a, b))

struct HTTP_Head_Line *http_head;

struct HTTP_Head_Line *HTTP_AddHeadLine(char *line)
{
	struct HTTP_Head_Line *h;

	translate(line, "\r\n", "\0\0");
	h = (struct HTTP_Head_Line *)gmalloc(sizeof(struct HTTP_Head_Line));

	h->line = (char *)gmalloc(strlen(line)+1);
	strcpy(h->line, line);

	if (!http_head) {
	  http_head = h;
	  h->next = 0;
	}
	else {
	  h->next = http_head;
	  http_head = h;
	}

	return(h);
}

void HTTP_ClearHeadLines(void)
{
	struct HTTP_Head_Line *h, *i;

    for (i = http_head; i; i = h) {
	  h = i->next;
	  free(i->line);
	  free(i);
	}
	http_head = 0;
}

char *HTTP_FindHeadLine(char *leftpart)
{
	struct HTTP_Head_Line *h;
	int l;

	l = strlen(leftpart);

	for (h = http_head; h; h = h->next) {
	  if (!str_nieq(h->line, leftpart, l)) return(h->line);
	}
	return 0;
}

void HTTP_PostConfig(void)
{
	http_head = 0;
}

/*
** grab->filename is not freed if filename or grab->uri->saveto
** are not defined
*/

#ifdef __SASC
static char odir[256];
#else
static int odir;
#endif

int GetHTTPFile(struct Grab *grab, struct URI *uri, char *filename)
{
	int ret, ReGetFile=0, doexists;
	char *p, tok[256];

	memset(grab, 0, sizeof(struct Grab));

	grab->uri = uri;
	grab->suri = SplitURI(uri->uri);

	if (filename) grab->filename = filename;
	else if (grab->uri->saveto) grab->filename = grab->uri->saveto;
	else if (grab->uri->flags & URI_NODIRS) {
	  /* We make a filename - no path */
	  char *tstr;

	  tstr = (grab->suri->filename ?
			  grab->suri->filename : cfg_global.def_filename);
	  grab->filename = (char *)gmalloc(strlen(tstr) + 1);
	  strcpy(grab->filename, tstr);
	}
	else {
	  /* We make a filename */
	  int len;

	  len = strlen(grab->suri->host) + 1;
	  if (grab->suri->path) len += strlen(grab->suri->path);
	  len += (grab->suri->filename ? strlen(grab->suri->filename)
			  : strlen(cfg_global.def_filename)) + 1;
	  grab->filename = (char *)gmalloc(len);
	  strcpy(grab->filename, grab->suri->host);
	  strcat(grab->filename, "/");
	  if (grab->suri->path) strcat(grab->filename, grab->suri->path);
	  strcat(grab->filename, (grab->suri->filename ? grab->suri->filename
							  : cfg_global.def_filename));
	}

	if (cfg_global.trans_from && !filename && !grab->uri->saveto) {
	  /* We "translate" the filename */
	  translate(grab->filename, cfg_global.trans_from, cfg_global.trans_to);
	}
	/****/
	/*	printf("Saving to '%s'\n", grab->filename); */

	sprintf(tok, "[%d/%d] %s%s", CountURI(URI_RECEIVED, -1)+1,
			CountURI(-1, URI_TYPES), grab->suri->nuri,
			(grab->uri->flags & URI_RETRY ? " [retry]" : ""));
	putlog(tok, LOG_TRANSFER);

#ifdef __SASC
	if (getcwd(odir, 256) == 0) {
#else
	odir = open(".", O_RDONLY);
	if (odir == -1) {
#endif
	  fprintf(stderr, "Error openning current directory\n");
	  exit(EXIT_FAILURE);
	}

	if (cfg_global.def_flags & URI_NONEXISTS) {
	  if (chdir(cfg_global.saveroot)) {
		fprintf(stderr, "GetHTTPFile: Can't CD to SaveRoot (\"%s\")\n",
				cfg_global.saveroot);
		exit(EXIT_FAILURE);
	  }

	  doexists = access(grab->filename, F_OK);
#ifdef __SASC
	  chdir(odir);
#else
	  fchdir(odir);
#endif
	  if (doexists == 0) {
		/* The file exists */
		putlog("Already on disk", LOG_TRANSFER);
		goto GetHTTPFile_RecdOK;
	  }
	}

 ReGetHTTPFile:

	ReGetFile = 0;

	ret = 0;

	if (HTTP_connect(grab)) {
	  if (HTTP_sendreq(grab)) {
		if (HTTP_gethead(grab)) {
		  if (grab->retcode == 200) {
			if (HTTP_getbody(grab)) {
			  /* Success */
			GetHTTPFile_RecdOK:
			  uri->flags &= ~URI_TYPES;
			  uri->flags |= URI_RECEIVED;
			  ret = 1;
			}
			else {
			  uri->flags &= ~URI_TYPES;
			  uri->flags |= URI_FAILED;
			}
		  }
		  else {
		    /* Retcode != 200 */
		    fprintf(stderr, "\033[K");
			if (grab->retcode == 301 || grab->retcode == 302) {
			  /* Moved */
			  p = HTTP_FindHeadLine("location: ");
			  if (p) {
				char *p2;
				char tok[128];

				p2 = ExtractArgument(p, 1);
				sprintf(tok, "Moved to '%s'", p2);
				putlog(tok, LOG_TRANSFER);
				AddURI(grab->uri, p2, -2, -2, grab->uri->saveto);
				uri->flags &= ~URI_TYPES;
				uri->flags |= URI_MOVED;
				free(p2);
			  }
			  else {
				putlog("Moved to unknown", LOG_TRANSFER);
				uri->flags &= ~URI_TYPES;
				uri->flags |= URI_MOVED;
			  }
			}
			else if (grab->retcode == 304) {
			  /* Not modified - mark received */
			  uri->flags &= ~URI_TYPES;
			  uri->flags |= URI_RECEIVED;
			  putlog("304 Not modified", LOG_TRANSFER);
			  ret = 1;
			}
			else if (grab->retcode == 401) {
			  /* Authentication error */
			  char *line;
			  line = HTTP_FindHeadLine("WWW-Authenticate: Basic");
			  if (!line) {
				putlog("Authentication error, unknown realm", LOG_TRANSFER);
			  }
			  else {
				  p = strchr(line, '=');
				  if (!p) {
					putlog("Authentication error, bad realm descriptor",
						   LOG_TRANSFER);
				  }
				  else {
					struct Cfg_Auth *cauth;
					char *p2;

					p++;
					p2 = strchr(p, '\r');
					if (p2) p2[0]=0;
					p2 = strchr(p, '\n');
					if (p2) p2[0]=0;

					p2 = Dequote(p);

					if (grab->authpwd) {
					  char tok[256];
					  sprintf(tok, "Authentication error, bad user/pwd for"
							  " realm '%s'", p2);
					  putlog(tok, LOG_TRANSFER);
					}
					else {
					  for (cauth = cfg_global.auth; cauth;
						   cauth = cauth->next) {
						if (!str_ieq(p2, cauth->realm)) {
						  /* match! */
						  grab->authpwd = cauth->upw;
						  ReGetFile = 1;
						}
					  }
					  if (!ReGetFile) {
						char tok[256];
						sprintf(tok, "Authentication error: unknown"
								" user/password for realm '%s'", p2);
						putlog(tok, LOG_TRANSFER);
					  }
					}
				  }
			  }
			  uri->flags &= ~URI_TYPES;
			  uri->flags |= URI_ERROR;
			}
			else if (grab->retcode == 403 || grab->retcode == 404) {
			  uri->flags &= ~URI_TYPES;
			  uri->flags |= URI_ERROR;
			  putlog("Forbidden/not found", LOG_TRANSFER);
			}
			else if (grab->retcode >= 500) {
			  uri->flags &= ~URI_TYPES;
			  uri->flags |= URI_ERROR;
			  putlog("Server error", LOG_TRANSFER);
			}
			else {
			  char tok[128];
			  uri->flags &= ~URI_TYPES;
			  uri->flags |= URI_ERROR;
			  sprintf(tok, "Unknown server return code: %d", grab->retcode);
			  putlog(tok, LOG_TRANSFER);
			}
		  }
		}
		else {
		  uri->flags &= ~URI_TYPES;
		  uri->flags |= URI_ERROR;
		}
	  }
	}
	else {
	  uri->flags &= ~URI_TYPES;
	  uri->flags |= URI_ERROR;
	}

	if (ReGetFile) {
	  HTTP_ClearHeadLines();
	  if (grab->sock) {
		close(grab->sock);
		grab->sock = 0;
	  }
	  goto ReGetHTTPFile;
	}

#ifndef __SASC
	close(odir);
#endif
	return(ret);
}

int lookup(struct sockaddr_in *ski, const char *host, int port)
{
	struct hostent *remote;

	remote = gethostbyname(host);
	if (!remote) {
	  /* Lookup failure */
	  return(0);
	}

	memcpy(&ski->sin_addr, remote->h_addr, sizeof(ski->sin_addr));
#ifdef NO_HTONS
	ski->sin_port = (port ? port : 80);
#else
	ski->sin_port = htons((port ? port : 80));
#endif
	ski->sin_family = AF_INET;
	return(1);
}

int HTTP_connect(struct Grab *grab)
{
	struct sockaddr_in ski;

	fprintf(stderr, "-> Looking up...\033[A\n");

	if (!lookup(&ski, grab->suri->host, grab->suri->port)) {
	  char tok[128];
	  sprintf(tok, "Can't find host '%s'", grab->suri->host);
	  putlog(tok, LOG_TRANSFER);
	  return 0;
	}

	fprintf(stderr, "\033[K-> Connecting...\033[A\n");

	grab->sock = socket(AF_INET, SOCK_STREAM, 0);
	if (grab->sock == -1) {
	  /* Error */
	  fprintf(stderr, "HTTP_connect: Can't create socket\n");
	  exit(EXIT_FAILURE);
	}

	/* Set non-blocking mode */
	/*
	flgs = fcntl(grab->sock, F_GETFL, 0);
	fcntl(grab->sock, flgs | O_NONBLOCK);
	*/
	if (connect(grab->sock, (struct sockaddr *)(&ski),
				sizeof(struct sockaddr_in)) < 0) {
	  /* Error - see man perror */
	  char tok[128];
	  sprintf(tok, "Can't connect: %s", strerror(errno));
	  putlog(tok, LOG_TRANSFER);
	  return(0);
	}
	return(1);
}

char *wkday[7] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
char *month[12] = { "Jan", "Feb", "Mar", "apr", "May", "Jun",
					"Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

int HTTP_sendreq(struct Grab *grab)
{
	time_t now;
	struct tm *gmt;
	char cp[256];
	struct stat finfo;
	struct RequestLine *rl;

	/* Test socket */
	/*	send(grab->sock, 0, 0, 0); */
	/*	if (errno) { */
	  /* Socket not OK.. */
	/*	  return(0); */
	/*	} */

	fprintf(stderr, "\033[K-> Sending request...\033[A\n");

	/* Send GetMesg */
	send(grab->sock, "GET /", 5, 0);
	send(grab->sock, grab->suri->path,
		 (grab->suri->path ? strlen(grab->suri->path) : 0), 0);
	if (grab->suri->filename)
	  send(grab->sock, grab->suri->filename, strlen(grab->suri->filename), 0);
	send(grab->sock, " HTTP/1.0\r\n", 11, 0);

	/* Partial HTTP/1.1 support */
	send(grab->sock, "Host: ", 6, 0);
	send(grab->sock, grab->suri->host, strlen(grab->suri->host), 0);
	send(grab->sock, "\r\n", 2, 0);

	/* We only support one file/connection for now */
	send(grab->sock, "Connection: close\r\n", 19, 0);

	/* General header: RFC1123 date and pragma */
	now = time(0);
	gmt = gmtime(&now);
	sprintf(cp, "%s, %02u %s %04u %02u:%02u:%02u GMT\r\n",
		wkday[gmt->tm_wday], gmt->tm_mday, month[gmt->tm_mon],
		gmt->tm_year + 1900, gmt->tm_hour, gmt->tm_min, gmt->tm_sec);
	send(grab->sock, "Date: ", 6, 0);
	send(grab->sock, cp, strlen(cp), 0);
	/* No pragma to send :| */

	/* Request-header: Authorization, From, If-Modified-Since,
	   Referer, User-Agent */
	/* Authorization */
	if (grab->authpwd) {
	  send(grab->sock, "Authorization: Basic ", 21, 0);
	  send(grab->sock, grab->authpwd, strlen(grab->authpwd), 0);
	  send(grab->sock, "\r\n", 2, 0);
	}
	/* From */
	if (cfg_global.fromstring) {
	  send(grab->sock, "From: ", 6, 0);
	  send(grab->sock, cfg_global.fromstring,
		   strlen(cfg_global.fromstring), 0);
	  send(grab->sock, "\r\n", 2, 0);
	}
	/* If-Modified-Since */
	if (str_ieq(grab->filename, "-") &&
	    (grab->uri->flags & URI_IFMODIFIED) &&
	    !chdir(cfg_global.saveroot)) {
	  if (stat(grab->filename, &finfo) == 0) {
	    gmt = gmtime(&(finfo.st_mtime));
	    sprintf(cp, "%s, %02u %s %04u %02u:%02u:%02u GMT\r\n",
		    wkday[gmt->tm_wday], gmt->tm_mday, month[gmt->tm_mon],
		    gmt->tm_year + 1900, gmt->tm_hour, gmt->tm_min, gmt->tm_sec);
	    send(grab->sock, "If-Modified-Since: ", 19, 0);
	    send(grab->sock, cp, strlen(cp), 0);
	  }
#ifdef __SASC
	  chdir(odir);
#else
	  fchdir(odir);
#endif
	}
	/* Referer unsupported */
	/* User-Agent */
	send(grab->sock, "User-Agent: ", 12, 0);
	send(grab->sock, gu_version, strlen(gu_version), 0);
	send(grab->sock, "\r\n", 2, 0);

	/* Additional header fields from config-file */
	for (rl = cfg_global.shead; rl; rl = rl->next) {
	  send(grab->sock, rl->line, strlen(rl->line), 0);
	  send(grab->sock, "\r\n", 2, 0);
	}

	/* End of header */
	send(grab->sock, "\r\n", 2, 0);

	return(1);
}

#define HTTP_GETH_LINELEN 1024

int HTTP_gethead(struct Grab *grab)
{
	int rread, ret;
	char buf[HTTP_GETH_LINELEN];
	char *p;
	FILE *diskfile=0;

	fprintf(stderr, "\033[K-> Receiving header...\033[A\n");

	rread = fdgetdelim(grab->sock, buf, HTTP_GETH_LINELEN, "\r", "\n");
	if (rread <= 0) {
	  /* Error.. */
	  putlog("Can't get header", LOG_TRANSFER);
	  return(0);
	}

	ret = 1;
	if (cfg_global.def_flags & URI_SAVEHEAD) {
	  if (!str_ieq(grab->filename, "-")) {
		/* To stdout */
		diskfile = stdout;
	  }
	  else {
		if (chdir(cfg_global.saveroot)) {
		  fprintf(stderr, "HTTP_gethead: Can't cd to SaveRoot\n");
		  exit(EXIT_FAILURE);
		}

		p = (char *)gmalloc(strlen(grab->filename) + 5);
		strcpy(p, grab->filename);
		strcat(p, ".hdr");

		if (magic_makedirs(p, cfg_global.d_mode)) {
		  fprintf(stderr, "HTTP_gethead: Can't create directories leading"
				          " to \"%s\"\n", p);
		  exit(EXIT_FAILURE);
		}

		diskfile = fopen(p, "w");
		chmod(p, cfg_global.f_mode);
#ifdef __SASC
		chdir(odir);
#else
		fchdir(odir);
#endif
		free(p);
		if (!diskfile) {
		  fprintf(stderr, "HTTP_gethead: Can't open disk file\n");
		  exit(EXIT_FAILURE);
		}
	  }
	}

	if (str_nieq(buf, "HTTP/1.", 7)) {
	  /* Bad header? */
	  putlog("Bad header", LOG_TRANSFER);
	  ret = 0;
	}

	p = NextBlank(buf);
	p = NextNonBlank(p);
	/* p points to the error code/message */
	grab->retcode = strtol(p, 0, 10);

	HTTP_AddHeadLine(buf);
	if (diskfile) fprintf(diskfile, buf);

	while (1) {
	  rread = fdgetdelim(grab->sock, buf, HTTP_GETH_LINELEN, 0, "\n");
	  if (rread <= 0) {
		/* Error */
		putlog("Can't get header", LOG_TRANSFER);
		if (diskfile && diskfile != stdout) fclose(diskfile);
		return(0);
	  }

	  if (diskfile) fprintf(diskfile, buf);

	  if ((buf[0]=='\r' && buf[1]=='\n') || (buf[0]=='\n')) break;

	  HTTP_AddHeadLine(buf);

	  if (!str_nieq(buf, "content-length: ", 16)) {
		grab->size = atoi(buf+16);
		grab->size = strtol(buf+16, 0, 10);
	  }
	}

	  
	if (diskfile && diskfile != stdout) fclose(diskfile);

	return(1);
}

#define HTTP_BUF_SIZE 2048

int UN=1;

int HTTP_getbody(struct Grab *grab)
{
	FILE *diskfile;
	char buf[HTTP_BUF_SIZE], tok[256];
	int got, filestatus, timediff, cps;
	struct timeval tv;
	fd_set act_fds, read_fds;
	char i;
	int a;

	if (grab->size)
	  fprintf(stderr, "\033[K-> Receiving file (%u bytes)\033[A\n", grab->size);
	else
	  fprintf(stderr, "\033[K-> Receiving file\033[A\n");

	if (!str_ieq(grab->filename, "-")) {
	  diskfile = stdout;
	}
	else {
	  if (chdir(cfg_global.saveroot)) {
		fprintf(stderr, "HTTP_getbody: Can't CD to SaveRoot (\"%s\")\n",
				cfg_global.saveroot);
		exit(EXIT_FAILURE);
	  }
	  if (magic_makedirs(grab->filename, cfg_global.d_mode)) {
		fprintf(stderr, "HTTP_getbody: Can't create directories"
				" leading to \"%s\"\n",
				grab->filename);
		exit(EXIT_FAILURE);
	  }
	  diskfile = fopen(grab->filename, "w");
	  chmod(grab->filename, cfg_global.f_mode);
#ifdef __SASC
	  chdir(odir);
#else
	  fchdir(odir);
#endif
	  if (!diskfile) {
		fprintf(stderr, "HTTP_getbody: Can't open disk file\n");
		exit(EXIT_FAILURE);
	  }
	}

	ioctl(grab->sock, FIONBIO, (char *)&UN);

	grab->start = grab->ltouch = time(0);

	FD_ZERO(&act_fds);
#ifndef __SASC
	FD_SET(0, &act_fds);
#endif
	FD_SET(grab->sock, &act_fds);
	
	tv.tv_sec = 1; tv.tv_usec = 0;
	
	while(1) {
	  read_fds = act_fds;
	  if ((a = select(FD_SETSIZE, &read_fds, 0, 0, &tv)) < 0) {
printf("errno = %d, %x\n", errno, errno);
printf("EBADF: %d, EINTR: %d, EINVAL: %d\n", EBADF, EINTR, EINVAL);
printf("FD_SETSIZE = %d\n", FD_SETSIZE);
		fprintf(stderr, "HTTP_getbody: Select error\n");
		exit(EXIT_FAILURE);
	  }
	  if (!a) {
		tv.tv_sec = 1; tv.tv_usec = 0;
		timediff = grab->ltouch - grab->start;
		if (grab->size) {
		  if (timediff > 0) {
			cps = grab->recd / timediff;
			if (cps > 1023)
			  fprintf(stderr, "\033[K-> Receiving file [%u%%] [%u/%u bytes]"
					  " [%.1f Kbps - %u secs to go]\033[A\n",
					  100 * grab->recd / grab->size, grab->recd, grab->size,
					  (float)cps / 1024.0,
					  (grab->size - grab->recd) / cps + 1);
			else if (cps)
			  fprintf(stderr, "\033[K-> Receiving file [%u%%] [%u/%u bytes]"
					  " [%u bps - %u secs to go]\033[A\n",
					  100 * grab->recd / grab->size, grab->recd, grab->size,
					  cps, (grab->size - grab->recd) / cps + 1);
			else
			  fprintf(stderr, "\033[K-> Receiving file [%u%%] [%u/%u bytes]"
					  " [%u bps]\033[A\n",
					  100 * grab->recd / grab->size, grab->recd,
					  grab->size, cps);
			
		  }
		  else {
			fprintf(stderr, "\033[K-> Receiving file [%u%%] [%u/%u bytes]\033[A\n",
					100 * grab->recd / grab->size, grab->recd, grab->size);
		  }
		}
		else {
		  if (timediff > 0) {
			cps = grab->recd / timediff;
			if (cps > 1023)
			  fprintf(stderr, "\033[K-> Receiving file [%u bytes]"
					  " [%.1f Kbps]\033[A\n",
					  grab->recd, (float)cps / 1024.0);
			else
			  fprintf(stderr, "\033[K-> Receiving file [%u bytes]"
					  "[%u bps]\033[A\n",
					  grab->recd, cps);
		  }
		  else {
			fprintf(stderr, "\033[K-> Receiving file [%u bytes]\033[A\n",
					grab->recd);
		  }
		}
	  }
	  if (FD_ISSET(0, &read_fds)) {
		read(0, &i, 1);
		if (i=='q' || i=='Q') {
		  filestatus='q';
		  break;
		}
		else if (i=='z' || i=='Z') {
		  filestatus='z';
		  break;
		}
	  }
	  if (FD_ISSET(grab->sock, &read_fds)) {
	  ReadBodyFromSocket:
		got = read(grab->sock, buf, HTTP_BUF_SIZE);
		if (got > 0) {
		  if (fwrite(buf, 1, got, diskfile) != got) {
			/* Write error */
			fprintf(stderr, "HTTP_getbody: Error writing to file\n");
			exit(EXIT_FAILURE);
		  }
		  grab->recd += got;
		  if (got == HTTP_BUF_SIZE) goto ReadBodyFromSocket;
		  grab->ltouch = time(0);
		}
		else if (got == 0) {
		  /* EOF */
		  filestatus='e';
		  break;
		}
		else if (errno!=EAGAIN) {
		  fprintf(stderr, "HTTP_getbody: Errro reading from socket\n");
		  if (diskfile != stdout) fclose(diskfile);
		  exit(EXIT_FAILURE);
		}
	  }
	}

	if (diskfile != stdout) fclose(diskfile);

	grab->end = grab->ltouch = time(0);

	if (filestatus=='q') {
	  putlog("Quit by user", LOG_TRANSFER);
	  exit(0);
	}
	else if (filestatus=='z') {
	  putlog("File transfer abort by user", LOG_TRANSFER);
	  return(0);
	}

	timediff = grab->end - grab->start;

	if (timediff > 2) {
	  cps = grab->recd/timediff;
	  if (cps > 1023)
		sprintf(tok, "Received %u bytes in %u secs (%.1f Kbps)",
				grab->recd, timediff,
				(float)cps/1024.0);
	  else
		sprintf(tok, "Received %u bytes in %u secs (%u bps)",
				grab->recd, timediff, cps);
	}
	else {
	  sprintf(tok, "Received %u bytes in %u sec%s", grab->recd, timediff,
			  (timediff > 1 ? "s" : ""));
	}
	putlog(tok, LOG_TRANSFER);

	return(1);
}
